"""State module for managing EC2 VPC Peering Connection Accepter."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str,
    connection_status: str = None,
) -> Dict[str, Any]:
    """Requests to modify the VPC peering connection accepter.

    If you are creating a VPC peering connection in the same AWS account, you must both create and accept the request yourself.
    A VPC peering connection that's in the pending-acceptance state must be accepted by the owner of the accepter VPC to be activated.
    Updating the status only makes sense when the vpc peering connection is in pending-acceptance mode.
    When updating the status is attempted to be made through the VPC peering connection resource,
    it can only take place when the two VPCs are in the same region and have the same owner.
    If that is not true, the acceptance is controlled by this particular resource - VPC peering connection accepter.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str):
            An identifier of the resource in the provider. It is an identifier of the
            vpc peering connection of which the options are part of.

        connection_status(str, Optional):
            The desired status for the VPC peering connection, but in practice,
            an update will be attempted only if this status is set to "active".

    Request Syntax:
        .. code-block:: sls

            [vpc-peering-connection-id]:
              aws.ec2.vpc_peering_connection_options.present:
                - resource_id: "string"
                - connection_status: "string"

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            pcx-ae89ce9b:
              aws.ec2.vpc_peering_connection_options.present:
                - resource_id: pcx-ae89ce9b
                - connection_status: "active"
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    if not resource_id:
        result["result"] = False
        result[
            "comment"
        ] = hub.tool.aws.comment_utils.invalid_parameter_provided_comment(
            "resource_id", "aws.ec2.vpc_peering_connection_accepter", name
        )
        return result
    else:
        before = await hub.exec.aws.ec2.vpc_peering_connection_accepter.get(
            ctx, name=name, resource_id=resource_id
        )

        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = before["ret"]
        plan_state = copy.deepcopy(result["old_state"])

        # Update status - activate the connection
        # The only update possible is status activation - accepting the connection.
        # All the other statuses a connection could have are changed on their own,
        # for example: failed, expired, rejected...
        # The deleted status could be set by calling the absent function on this resource.
        if (
            connection_status == "active"
            and connection_status != result["old_state"]["connection_status"]
        ):
            accept_ret = await hub.tool.aws.ec2.vpc_peering_connection_utils.activate_vpc_peering_connection(
                ctx,
                "aws.ec2.vpc_peering_connection_accepter",
                name,
                resource_id,
                before["ret"]["connection_status"],
                connection_status,
            )

        result["result"] = accept_ret["result"]
        result["comment"] += accept_ret["comment"]
        if not accept_ret["result"]:
            return result

        resource_updated = resource_updated or bool(accept_ret["ret"])

        if resource_updated and ctx.get("test", False):
            plan_state["connection_status"] = accept_ret["ret"]["connection_status"]

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    "aws.ec2.vpc_peering_connection_accepter", name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    "aws.ec2.vpc_peering_connection_accepter", name
                )
        else:
            result["comment"] += hub.tool.aws.comment_utils.already_exists_comment(
                "aws.ec2.vpc_peering_connection_accepter", name
            )

        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not result["old_state"]) or resource_updated:
            after = await hub.exec.aws.ec2.vpc_peering_connection_accepter.get(
                ctx=ctx, name=name, resource_id=resource_id
            )
            if not after["result"]:
                result["result"] = False
                result["comment"] += after["comment"]
                return result

            result["new_state"] = copy.deepcopy(after["ret"])
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    """A No-Op function for vpc_peering_connection_accepter.

    This is a configuration resource of the vpc_peering_connection resource.
    It's not possible to delete vpc_peering_connection_accepter,
    You need to delete the vpc_peering_connection resource by calling vpc_peering_connection.absent,
    while providing the vpc_peering_connection id.
    If you want to modify the vpc_peering_connection_accepter resource,
    use the vpc_peering_connection_options.present.

    Args:
        name:
            An Idem name of the resource.

        resource_id:
            The id of the vpc peering connection.

    Request Syntax:
        .. code-block:: sls

            [vpc-peering-connection-id]:
              aws.ec2.vpc_peering_connection_accepter.absent:
                - name: "string"

    Returns:
        Dict[str, Any]
    """
    result = dict(
        comment=(
            "No-Op: The VPC peering connection accepter can not be deleted as it is not an AWS resource",
        ),
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Describes one or more of your VPC peering connection accepters.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.ec2.vpc_peering_connection_accepter.py
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_vpc_peering_connections(ctx)

    if not ret["result"]:
        hub.log.debug(
            f"Could not describe aws.ec2.vpc_peering_connection {ret['comment']}"
        )
        return {}

    for vpc_peering_connection in ret["ret"]["VpcPeeringConnections"]:
        resource_id = vpc_peering_connection.get("VpcPeeringConnectionId")
        vpc_peering_connection_accepter_translated = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_peering_connection_accepter_to_present(
            raw_resource=vpc_peering_connection, idem_resource_name=resource_id
        )

        result[vpc_peering_connection_accepter_translated["resource_id"]] = {
            "aws.ec2.vpc_peering_connection_accepter.present": [
                vpc_peering_connection_accepter_translated
            ]
        }

    return result



"""Exec module for managing VPC Peering Connection."""
from typing import Dict
from typing import List


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    filters: List = None,
) -> Dict:
    """Get EC2 vpc peering connection options from AWS account.

    Use an un-managed VPC peering connection as a data-source.
    Supply one of the inputs as the filter.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str, Optional):
            AWS VPC peering connection id to identify the resource.

        filters(list, Optional):
            One or more filters: for example, tag :<key>, tag-key.
            A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_vpc_peering_connections

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            $ idem exec aws.ec2.vpc_peering_connection.get

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.vpc_peering_connection.get
                - kwargs:
                    name: some-idem-resource-name
                    resource_id: some-resource-id
                    filters:
                      - name: 'vpc-peering-connection-id'
                        values: ["pcx-ae89ce9b"]
                      - name: 'tag:Name'
                        values: ["my_unmanaged_resource"]
    """
    result = await hub.tool.aws.ec2.vpc_peering_connection_utils.get_vpc_peering_connection_raw(
        ctx, name, resource_id, "aws.ec2.vpc_peering_connection", filters
    )
    raw_resource = result["ret"]

    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_peering_connection_to_present(
        raw_resource=raw_resource, idem_resource_name=name
    )

    return result



"""State module for managing Route53 Hosted zones."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    hosted_zone_name: str,
    caller_reference: str,
    resource_id: str = None,
    vpcs: List[
        make_dataclass(
            "VPC",
            [
                ("VPCRegion", str, field(default=None)),
                ("VPCId", str, field(default=None)),
            ],
        )
    ] = None,
    config: make_dataclass(
        "HostedZoneConfig",
        [
            ("Comment", str, field(default=None)),
            ("PrivateZone", bool, field(default=None)),
        ],
    ) = None,
    delegation_set_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass("Tag", [("Key", str), ("Value", str, field(default=None))])
    ] = None,
) -> Dict[str, Any]:
    """Creates a new public or private hosted zone.

    You create records in a public hosted zone to define how you want to route traffic on the internet for a domain, such
    as example.com, and its subdomains (apex.example.com, acme.example.com). You create records in a private hosted zone
    to define how you want to route traffic for a domain and its subdomains within one or more Amazon Virtual Private Clouds
    (Amazon VPCs). You can't convert a public hosted zone to a private hosted zone or vice versa. Instead, you must
    create a new hosted zone with the same name and create new resource record sets.  For more information about charges
    for hosted zones, see Amazon Route 53 Pricing.

    Note the following: You can't create a hosted zone for a top-level domain (TLD) such as .com.   For public hosted
    zones, Route 53 automatically creates a default SOA record and four NS records for the zone. For more information
    about SOA and NS records, see NS and SOA Records that Route 53 Creates for a Hosted Zone in the Amazon Route 53
    Developer Guide. If you want to use the same name servers for multiple public hosted zones, you can optionally
    associate a reusable delegation set with the hosted zone. See the DelegationSetId element.   If your domain is
    registered with a registrar other than Route 53, you must update the name servers with your registrar to make Route 53
    the DNS service for the domain. For more information, see Migrating DNS Service for an Existing Domain to Amazon
    Route 53 in the Amazon Route 53 Developer Guide. When you submit a CreateHostedZone request, the initial status of
    the hosted zone is PENDING. For public hosted zones, this means that the NS and SOA records are not yet available on
    all Route 53 DNS servers. When the NS and SOA records are available, the status of the zone changes to INSYNC. The
    CreateHostedZone request requires the caller to have an ec2:DescribeVpcs permission.

    Args:
        name(str):
            An Idem name of the hosted zone resource.

        hosted_zone_name(str):
            A unique string that identifies a hosted zone.

        caller_reference(str):
            A unique string that identifies the request and that allows failed CreateHostedZone
            requests to be retried without the risk of executing the operation twice.
            You must use a unique CallerReference string every time you submit a CreateHostedZone request.
            CallerReference can be any unique string, for example, a date/time stamp.,

        resource_id(str, Optional):
            AWS route53 hosted zone ID.

        config(dict[str, Any], Optional):
            A complex type that contains the following optional values:   For public and private
            hosted zones, an optional comment   For private hosted zones, an optional PrivateZone element
            If you don't specify a comment or the PrivateZone element, omit HostedZoneConfig and the other
            elements. Defaults to None.

            * Comment (str, Optional):
                Any comments that you want to include about the hosted zone.
            * PrivateZone (bool, Optional):
                A value that indicates whether this is a private hosted zone.

        vpcs(list, Optional):
            (Private hosted zones only) A list that contains information about the Amazon VPC's that
            you're associating with this hosted zone. If you are associating VPC's with a hosted zone with this request, the
            paramaters VPCId and VPCRegion are also required. To associate additional Amazon VPCs with the
            hosted zone, use AssociateVPCWithHostedZone after you create a hosted zone. Defaults to None.

            * VPCRegion (str, Optional):
                (Private hosted zones only) The region that an Amazon VPC was created in.
            * VPCId (str, Optional):
                (Private hosted zones only) The ID of an Amazon VPC.

        delegation_set_id(str, Optional):
            Id of a reusable delegation set,

        tags(list or dict, Optional):
            list of tags in the format of ``[{"Key": tag-key, "Value": tag-value}]`` or dict in the format of
            ``{tag-key: tag-value}`` The tags to assign to the hosted zone. Defaults to None.

            * Key (str):
                The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
            * Value (str):
                The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.


    Request Syntax:
        .. code-block:: sls

            [hosted_zone_id]:
              aws.route53.hosted_zone.present:
                - name: 'string'
                - hosted_zone_name: 'string'
                - resource_id: 'string'
                - caller_reference: 'string'
                - config:
                    PrivateZone: 'string'
                    Comment: 'string'
                - vpcs:
                  - VPCId: 'string'
                    VPCRegion: 'string'
                - delegation_set_id: 'string'
                - tags:
                    - Key: 'string'
                      Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.route53.hosted_zone.present:
                - name: value
                - hosted_zone_name: value
                - resource_id: value
                - caller_reference: value
                - config:
                    PrivateZone: True
                    Comment: 'description of hosted zone config'
                - vpcs:
                    - VPCId: 'vpc id'
                      VPCRegion: 'us-east1'
                - delegation_set_id: 'string'
                - tags:
                    - Key: 'tag key'
                      Value: 'tag value'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.route53.hosted_zone.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = copy.deepcopy(before["ret"])
        result["comment"] = (f"aws.route53.hosted_zone '{name}' already exists.",)
        plan_state = copy.deepcopy(result["old_state"])
        # Check if comments needs to be updated.
        if (
            config is not None
            and config.get("Comment")
            and config.get("Comment") != plan_state["config"].get("Comment")
        ):
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update aws.route53.hosted_zone '{name}'",
                )
                resource_updated = True
                plan_state["config"] = config
            else:
                update_ret_comment = (
                    await hub.exec.boto3.client.route53.update_hosted_zone_comment(
                        ctx,
                        Id=resource_id,
                        Comment=config.get("Comment"),
                    )
                )
                if update_ret_comment["result"]:
                    resource_updated = True
                    comment = config.get("Comment")
                    result["comment"] = result["comment"] + (
                        f"Updated comment to '{comment}'",
                    )
                else:
                    result["comment"] = update_ret_comment["comment"]
                    result["result"] = False
                    return result

        # Check if tags needs to be updated
        if tags is not None:
            update_ret_tags = await hub.tool.aws.route53.tag.update_tags(
                ctx,
                resource_type="hostedzone",
                resource_id=resource_id,
                old_tags=result["old_state"]["tags"],
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret_tags["comment"]
            if not update_ret_tags["result"]:
                result["result"] = False
                return result
            resource_updated = resource_updated or bool(update_ret_tags["ret"])
            if ctx.get("test", False) and update_ret_tags["ret"]:
                plan_state["tags"] = update_ret_tags["ret"]
    else:
        tags_dict = None
        if tags is not None:
            if isinstance(tags, List):
                tags_dict = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
            else:
                tags_dict = tags
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "hosted_zone_name": hosted_zone_name,
                    "caller_reference": caller_reference,
                    "config": config,
                    "vpcs": vpcs,
                    "delegation_set_id": delegation_set_id,
                    "tags": tags_dict,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.route53.hosted_zone", name=name
            )
            return result

        # Private Route53 Hosted Zones can only be created with their first VPC association,
        # however we need to associate the remaining after creation.

        vpc = None
        if vpcs is not None and len(vpcs) > 0:
            vpc = vpcs[0]

        ret = await hub.exec.boto3.client.route53.create_hosted_zone(
            ctx,
            Name=hosted_zone_name,
            VPC=vpc,
            CallerReference=caller_reference,
            HostedZoneConfig=config,
            DelegationSetId=delegation_set_id,
        )
        if not ret["result"]:
            result["result"] = False
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )
        resource_id = ret["ret"]["HostedZone"]["Id"].split("/")[-1]
        if vpcs is not None and len(vpcs) > 1:
            for i in range(1, len(vpcs)):
                vpc_ret = (
                    await hub.exec.boto3.client.route53.associate_vpc_with_hosted_zone(
                        ctx, HostedZoneId=resource_id, VPC=vpcs[i]
                    )
                )
                if not vpc_ret["result"]:
                    result["result"] = False
                    result["comment"] = vpc_ret["comment"]
                    hub.log.debug(
                        f"Could not associate vpc with hosted zone {vpc_ret['comment']}"
                    )
                    return result
        if tags is not None:
            add_tags = await hub.tool.aws.route53.tag.update_tags(
                ctx,
                resource_type="hostedzone",
                resource_id=resource_id,
                old_tags=None,
                new_tags=tags_dict,
            )
            if not add_tags["result"]:
                error = add_tags["comment"]
                result["result"] = False
                hub.log.debug(
                    f"Tag addition failed for hosted_zone {name} with error {error}"
                )
                result["comment"] = result["comment"] + error
                return result

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif result["old_state"] is None or resource_updated:
        after = await hub.exec.aws.route53.hosted_zone.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"]:
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Deletes a hosted zone.

    If the hosted zone was created by another service, such as Cloud Map, see Deleting Public Hosted Zones That Were
    Created by Another Service in the Amazon Route 53 Developer Guide for information about how to delete it.
    (The process is the same for public and private hosted zones that were created by another service.) If you want to
    keep your domain registration but you want to stop routing internet traffic to your website or web application, we
    recommend that you delete resource record sets in the hosted zone instead of deleting the hosted zone.  If you delete
    a hosted zone, you can't undelete it. You must create a new hosted zone and update the name servers for your domain
    registration, which can require up to 48 hours to take effect. (If you delegated responsibility for a subdomain to a
    hosted zone and you delete the child hosted zone, you must update the name servers in the parent hosted zone.)

    In addition, if you delete a hosted zone, someone could hijack the domain and route traffic to their own resources using
    your domain name.  If you want to avoid the monthly charge for the hosted zone, you can transfer DNS service for the
    domain to a free DNS service. When you transfer DNS service, you have to update the name servers for the domain
    registration. If the domain is registered with Route 53, see UpdateDomainNameservers for information about how to replace
    Route 53 name servers with name servers for the new DNS service. If the domain is registered with another registrar,
    use the method provided by the registrar to update name servers for the domain registration.

    For more information, perform an internet search on "free DNS service." You can delete a hosted zone only if it
    contains only the default SOA record and NS resource record sets. If the hosted zone contains other resource record
    sets, you must delete them before you can delete the hosted zone. If you try to delete a hosted zone that contains
    other resource record sets, the request fails, and Route 53 returns a HostedZoneNotEmpty error. For information about
    deleting records from your hosted zone, see ChangeResourceRecordSets. To verify that the hosted zone has been deleted,
    do one of the following:   Use the GetHostedZone action to request information about the hosted zone.   Use the
    ListHostedZones action to get a list of the hosted zones associated with the current Amazon Web Services
    account.

    Args:
        name(str):
            An Idem name of the hosted zone resource.
        resource_id(str, Optional):
            AWS route53 hosted zone ID. Idem automatically considers this resource being absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [hosted-zone-name]:
              aws.route53.hosted_zone.present:
                - name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.route53.hosted_zone.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )
        return result
    before = await hub.exec.aws.route53.hosted_zone.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.route53.delete_hosted_zone(
            ctx, Id=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.route53.hosted_zone", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Retrieves a list of the public and private hosted zones that are associated with the current Amazon Web Services
    account.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.route53.hosted_zone
    """

    result = {}

    ret = await hub.exec.boto3.client.route53.list_hosted_zones(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe hosted_zone {ret['comment']}")
        return result

    for hosted_zone in ret["ret"]["HostedZones"]:
        resource_id = hosted_zone.get("Id")
        name = hosted_zone.get("Name")
        resource_id = resource_id.split("/")[-1]
        resource = await hub.exec.aws.route53.hosted_zone.get(
            ctx, name=name, resource_id=resource_id
        )

        if not resource["result"]:
            hub.log.info(
                f"Failed to retrieve aws.route53.hosted_zone {name}: {resource['comment']}"
            )
            continue

        result[resource_id] = {
            "aws.route53.hosted_zone.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource["ret"].items()
            ]
        }

    return result


async def search(
    hub,
    ctx,
    name: str,
    hosted_zone_name: str = None,
    resource_id: str = None,
    private_zone: bool = None,
    vpc_id: str = None,
    tags: List = None,
):
    """
    Use an un-managed hosted_zone as a data-source. Supply one or more of the inputs as the filter.
    This function has been deprecated. Please use exec.run with aws.route53.hosted_zone.get instead.

    Args:
        name(str):
            The name of the Idem state.

        hosted_zone_name(str, Optional):
            The domain name of hosted_zone

        resource_id(str, Optional):
            AWS hosted_zone id to filter the resource.If resource_id is mentioned then we directly return the resource
            without checking other filters

        private_zone(bool, Optional):
            Bool argument to specify a private hosted_zone. One of the filter option for hosted_zone

        vpc_id(str, Optional):
            The vpc_id associated with the hosted_zone.One of the filter option for hosted_zone

        tags(list, Optional):
            Tags of the hosted_zone. One of the filter option for hosted_zone


    Request Syntax:
        .. code-block:: sls

            [Idem-state-name]:
              aws.route53.hosted_zone.search:
                - hosted_zone_name: 'string'
                - resource_id: 'string'
                - private_zone: 'bool'
                - vpc_id: 'string'
                - tags: 'List'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            my-unmanaged-hosted_zone:
              aws.route53.hosted_zone.search:
                - hosted_zone_name: idem-test-example-86a1d12e-4d68-4979-96ff-8608c33e4b9f.com.
                - resource_id: Z02905093MT09M4X5AHR5
                - private_zone: false
                - vpc_id: vpc-0e87ce8f2c3317d53
                - tags:
                    - Key: Name
                      Value: idem-test-hosted_zone

    """

    hub.log.warning(
        f"aws.route53.hosted_zone.search '{name}' state has been deprecated. Please use exec.run with aws.route53.hosted_zone.get instead."
    )
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if resource_id:
        # If resource_id is mentioned directly returning the hosted_zone
        ret = await hub.exec.aws.route53.hosted_zone.get(
            ctx, name=name, resource_id=resource_id
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = (
                f"Could not find hosted_zone with given {resource_id} resource_id {ret['comment']}",
            )
            return result
        result["comment"] = (
            f"Found hosted_zone with given resource_id {resource_id}, other filters will not be considered",
        )
        result["old_state"] = ret["ret"]
        # Populate both "old_state" and "new_state" with the same data
        result["new_state"] = copy.deepcopy(result["old_state"])
    else:
        # Get all the hosted_zones to apply the filters
        describe_ret = (
            await hub.tool.aws.route53.hosted_zone_utils.get_all_hosted_zones(ctx)
        )
        if not describe_ret["result"]:
            result["result"] = False
            result["comment"] = (
                f"Could not describe all the aws.route53.hosted_zone resources to apply the filters, {describe_ret['comment']}",
            )
            return result

        ret = hub.tool.aws.route53.hosted_zone_utils.get_hosted_zones_with_filters(
            raw_hosted_zones=describe_ret["ret"],
            hosted_zone_name=hosted_zone_name,
            private_zone=private_zone,
            vpc_id=vpc_id,
            tags=tags,
        )
        result["comment"] = result["comment"] + ret["comment"]
        if not ret["ret"]:
            result["result"] = ret["result"]
            return result

        # For backward compatibility since search returns a single object
        hosted_zone = ret["ret"][0]
        tags = hosted_zone.get("tags")

        result[
            "old_state"
        ] = hub.tool.aws.route53.conversion_utils.convert_raw_hosted_zone_to_present(
            raw_resource=hosted_zone,
            idem_resource_name=name,
            tags=tags,
        )
        # Populate both "old_state" and "new_state" with the same data
        result["new_state"] = copy.deepcopy(result["old_state"])

    return result



"""State module for managing Amazon API Gateway Integration Responses."""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    http_method: str,
    status_code: str,
    parent_resource_id: str,
    resource_id: str = None,
    response_templates: Dict = None,
    response_parameters: Dict = None,
    content_handling: str = None,
    selection_pattern: str = None,
):
    """
    Creates a new API Gateway Integration Response or modifies an existing one.

    Args:
        name(str):
            An idem name of the resource.

        rest_api_id(str):
            AWS rest_api id of the associated RestApi.

        http_method(str):
            String that specifies the method request's HTTP method type.

        status_code(str):
            The method response's status code.

        parent_resource_id(str):
            The parent resource's id.

        resource_id(str, Optional):
            Defaults to None. Idem Resource id, formatted as: rest_api_id-parent_resource_id-http_method-status_code.

        response_parameters(dict, Optional):
            A key-value map specifying response parameters that are passed to the method response from the back end.
                The key is a method response header parameter name and the mapped value is an integration response header
                value, a static value enclosed within a pair of single quotes, or a JSON expression from the integration
                response body. The mapping key must match the pattern of method.response.header.{name} , where name is a
                valid and unique header name. The mapped non-static value must match the pattern of
                integration.response.header.{name} or integration.response.body.{JSON-expression} , where name must be a
                valid and unique response header name and JSON-expression a valid JSON expression without the $ prefix.
                Must be of format {'string': 'string'}.

        response_templates(dict, Optional):
            Specifies a put integration response's templates. Must be of format {'string': 'string'}.

        content_handling(str, Optional):
            Specifies how to handle response payload content type conversions. Supported values are CONVERT_TO_BINARY
                and CONVERT_TO_TEXT , with the following behaviors: If this property is not defined, the response payload
                will be passed through from the integration response to the method response without modification.

        selection_pattern(str, Optional):
            Specifies the selection pattern of a put integration response.

    Returns:
        dict[str, Any]

    Request Syntax:
        .. code-block:: sls

            [idem_test_aws_apigateway_integration_response]:
              aws.apigateway.integration_response.present:
                - name: 'string'
                - rest_api_id: 'string'
                - parent_resource_id: 'string'
                - http_method: 'string'
                - status_code: 'string'
                - response_parameters: dict
                - response_templates: dict
                - content_handling: 'string'
                - selection_pattern: 'string'

    Examples:
        .. code-block:: sls

            idem_test_apigateway_integration_response:
              aws.apigateway.integration_response.present:
                - name: 'idem_test_apigateway_integration_response'
                - rest_api_id: 'integration.get('rest_api_id')'
                - parent_resource_id: 'integration.get('parent_resource_id')'
                - http_method: 'GET'
                - status_code: '400'

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.apigateway.integration_response.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = before["ret"]

        update_parameters = {
            "response_templates": response_templates,
            "response_parameters": response_parameters,
            "content_handling": content_handling,
            "selection_pattern": selection_pattern,
        }

        update_integration_response_ret = (
            await hub.tool.aws.apigateway.integration_response.update(
                ctx,
                old_state=result["old_state"],
                updateable_parameters=update_parameters,
            )
        )
        result["comment"] = (
            result["comment"] + update_integration_response_ret["comment"]
        )

        if not update_integration_response_ret["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_integration_response_ret["ret"])

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "rest_api_id": rest_api_id,
                    "parent_resource_id": parent_resource_id,
                    "http_method": http_method,
                    "status_code": status_code,
                    "response_parameters": response_parameters,
                    "response_templates": response_templates,
                    "content_handling": content_handling,
                    "selection_pattern": selection_pattern,
                },
            )

            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.integration_response", name=name
            )
            return result

        resource_id = f"{rest_api_id}-{parent_resource_id}-{http_method}-{status_code}"

        ret = await hub.exec.boto3.client.apigateway.put_integration_response(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
            statusCode=status_code,
            responseTemplates=response_templates,
            responseParameters=response_parameters,
            contentHandling=content_handling,
        )

        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.integration_response", name=name
        )

    if (not before) or resource_updated:
        after = await hub.exec.aws.apigateway.integration_response.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result

        result["new_state"] = after["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None):
    r"""
    Deletes an API Gateway Integration Response.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            Idem Resource id, formatted as: rest_api_id-parent_resource_id-http_method-status_code. Defaults to None.
                Idem automatically considers this resource being absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

          [idem_test_aws_apigateway_integration_response]:
            aws.apigateway.integration_response.absent:
              - name: 'string'
              - resource_id: 'string'

    Returns:
        dict[str, Any]

    Examples:
        .. code-block:: sls

          idem_test_aws_apigateway_integration_response:
            aws.apigateway.integration_response.absent:
              - name: 'idem_test_apigateway_integration_response'
              - resource_id: 'rest_api_id-parent_resource_id-http_method-status_code'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        # if resource_id isn't specified, the resource is considered to be absent.
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.integration_response", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.integration_response.get(
        ctx,
        name=name,
        resource_id=resource_id,
    )

    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]

    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.integration_response", name=name
        )
        return result

    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.integration_response", name=name
        )
        return result

    else:
        rest_api_id, parent_resource_id, http_method, status_code = resource_id.split(
            "-"
        )
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigateway.delete_integration_response(
            ctx,
            restApiId=rest_api_id,
            resourceId=parent_resource_id,
            httpMethod=http_method,
            statusCode=status_code,
        )

        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.integration_response", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the API Gateway Integration Responses associated with a specific Integration.

    Returns a list of apigateway.integration_response descriptions

    Returns:
        dict[str, dict[str, Any]]


    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.integration_response

    """

    result = {}
    get_rest_apis_ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)
    if not get_rest_apis_ret["result"]:
        hub.log.debug(f"Could not get Rest Apis {get_rest_apis_ret['comment']}")
        return result

    for rest_api in get_rest_apis_ret["ret"]["items"]:
        rest_api_id = rest_api.get("id")

        get_resources_ret = await hub.exec.boto3.client.apigateway.get_resources(
            ctx, restApiId=rest_api_id
        )
        if not get_resources_ret["result"]:
            hub.log.debug(
                f"Could not get Resources {get_resources_ret['comment']}. Will skip this Resource."
            )
            continue

        if get_resources_ret["ret"]["items"] is not None:
            for resource in get_resources_ret["ret"]["items"]:
                parent_resource_id = resource.get("id")

                if resource.get("resourceMethods") is not None:
                    for resource_method in resource.get("resourceMethods"):
                        integration_resource_id = (
                            f"{rest_api_id}-{parent_resource_id}-{resource_method}"
                        )
                        integration_ret = await hub.exec.aws.apigateway.integration.get(
                            ctx, resource_id=integration_resource_id
                        )
                        if not integration_ret["result"]:
                            hub.log.debug(
                                f"Could not get Integration {integration_ret['comment']}. Will skip this Integration."
                            )
                            continue
                        integration_responses = integration_ret["ret"].get(
                            "integration_responses", []
                        )

                        for status_code in integration_responses:
                            int_resp_resource_id = f"{rest_api_id}-{parent_resource_id}-{resource_method}-{status_code}"

                            integration_response = (
                                await hub.exec.aws.apigateway.integration_response.get(
                                    ctx, resource_id=int_resp_resource_id
                                )
                            )
                            if not integration_response["result"]:
                                hub.log.debug(
                                    f"Could not get Integration Response {integration_response['comment']}. Will skip this Integration Response."
                                )
                                continue

                            resource_translated = integration_response["ret"]
                            result[resource_translated["resource_id"]] = {
                                "aws.apigateway.integration_response.present": [
                                    {parameter_key: parameter_value}
                                    for parameter_key, parameter_value in resource_translated.items()
                                ]
                            }
        return result



"""Exec module for assume role."""
from typing import Any
from typing import Dict
from typing import List


async def credentials(
    hub,
    ctx,
    role_arn: str,
    role_session_name: str,
    *,
    policy_arns: List[Dict[str, Any]] = None,
    policy: str = None,
    duration_seconds: int = None,
    tags: List[Dict[str, str]] = None,
    transitive_tag_keys: List[str] = None,
    external_id: str = None,
    serial_number: str = None,
    token_code: str = None,
    source_identity: str = None,
) -> Dict[str, Any]:
    """Returns a set of temporary security credentials to access AWS resources that you might not normally have access to.

    Args:
        role_arn(str):
            The ARN of the role to assume to access AWS resources.

        role_session_name(str):
            An identifier for the assumed role session. Use  the  role  session name to uniquely identify
            a session when the same role is assumed by different principals or for  different  reasons.

        policy_arns(list[dict[str, Any]], Optional):
            The  ARNs of the IAM managed policies that you want to use as managed session policies. The policies
            must exist in the same account as the role.

        policy(str, Optional):
            An  IAM policy in JSON format that you want to use as an inline session policy.

        duration_seconds(int, Optional):
            The  duration,  in seconds, of the role session. By default, the value is set to 3600 seconds.
            The value specified can range from 900 seconds (15 minutes) up to  the  maximum  session
            duration  which can have a value from 1 hour to 12 hours.

        tags(list[dict[str, str]], Optional):
            A list of session tags that you want to pass in the format of ``[{"Key": tag-key, "Value": tag-value}]``.
            Each session tag  consists  of  a  key name and an associated value. You can pass up to 50 session tags.
            The plaintext session tag keys can't exceed 128 characters, and the  values can't exceed 256 characters.

            * Key (*str*):
                The key of the tag.
            * Value (*str*):
                The value of the tag.

        transitive_tag_keys(list(str), Optional):
            A list of keys for session tags that you want to set as  transitive. If  you set a tag key as transitive,
            the corresponding key and value passes to subsequent sessions in a role chain.

        external_id(str, Optional):
            A unique identifier that might be required when you assume a role in another  account.

        serial_number(str, Optional):
            The identification number of the MFA device that is associated  with the  user  who  is
            making the credentials call. Specify this value if the trust policy of the role being assumed
            includes a condition that requires  MFA  authentication.

        token_code(str, Optional):
            The  value  provided  by  the MFA device, if the trust policy of the role being assumed requires MFA.

        source_identity(str, Optional):
            The source identity specified by the principal that is  calling  the credential operation.

    Returns:
        Dict[str, Any]:
            The temporary security credentials, which include an access key ID, a secret access key, and
            a security (or session) token.

    Examples:
        Call from the CLI:

        .. code-block:: bash

            $ idem exec aws.sts.assume_role.credentials <role_arn> <role_session_name>

        Call from code:

        .. code-block:: python

            async def my_func(hub, ctx, role_arn:str, role_session_name:str):
                await hub.exec.aws.sts.assume_role.credentials(ctx, role_arn, role_session_name)

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.sts.assume_role.credentials
                - kwargs:
                    role_arn: role_arn
                    role_session_name: role_session_name

    """
    ret = dict(result=True, ret={}, comment="")
    hub.log.debug(f"Assume role configuration set for ARN {role_arn}")

    config = {"RoleArn": role_arn, "RoleSessionName": role_session_name}
    if external_id is not None:
        config["ExternalId"] = external_id
    if policy_arns is not None:
        config["PolicyArns"] = policy_arns
    if policy is not None:
        config["Policy"] = policy
    if tags is not None:
        config["Tags"] = tags
    if transitive_tag_keys is not None:
        config["TransitiveTagKeys"] = transitive_tag_keys
    if duration_seconds is not None:
        config["DurationSeconds"] = duration_seconds
    if serial_number is not None:
        config["SerialNumber"] = serial_number
    if token_code is not None:
        config["TokenCode"] = token_code
    if source_identity is not None:
        config["SourceIdentity"] = source_identity

    assumed_role_object = await hub.exec.boto3.client.sts.assume_role(ctx, **config)
    ret["result"] = assumed_role_object.result
    ret["comment"] = assumed_role_object.comment
    if assumed_role_object.result:
        ret["ret"] = assumed_role_object.ret["Credentials"]
    return ret



""" State module for managing Kinesis Stream. """
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

ENCRYPTION_KMS_TYPE = "KMS"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    shard_count: int = None,
    stream_mode_details: make_dataclass(
        "StreamModeDetails", [("StreamMode", str)]
    ) = None,
    shard_level_metrics: List = None,
    retention_period_hours: int = None,
    encryption_type: str = None,
    key_id: str = None,
    stream_arn: str = None,
    tags: Dict[str, str] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=10)),
                        ("max_attempts", int, field(default=18)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=18)),
                        ("max_attempts", int, field(default=18)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates a Kinesis data stream. A stream captures and transports data records that are continuously emitted from
    different data sources or producers.

    Scale-out within a stream is explicitly supported by means of shards, which are uniquely identified groups of data
    records in a stream. You specify and control the number of shards that a stream is composed of. Each shard can
    support reads up to five transactions per second, up to a maximum data read total of 2 MiB per second. Each shard
    can support writes up to 1,000 records per second, up to a maximum data write total of 1 MiB per second. If the
    amount of data input increases or decreases, you can add or remove shards. The stream name identifies the stream.
    The name is scoped to the Amazon Web Services account used by the application. It is also scoped by Amazon Web
    Services Region. That is, two streams in two different accounts can have the same name, and two streams in the same
    account, but in two different Regions, can have the same name. CreateStream is an asynchronous operation. Upon
    receiving a CreateStream request, Kinesis Data Streams immediately returns and sets the stream status to CREATING.
    After the stream is created, Kinesis Data Streams sets the stream status to ACTIVE. You should perform read and
    write operations only on an ACTIVE stream. Enables enhanced Kinesis data stream monitoring for shard-level metrics.

    You receive a LimitExceededException when making a CreateStream request when you try to do one of the following:
        a. Have more than five streams in the CREATING state at any point in time.
        b. Create more shards than are authorized for your account.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        shard_count(int, Optional):
            The number of shards that the stream will use. The throughput of the stream is a function of the number of
            shards; more shards are required for greater provisioned throughput. Defaults to None.

        stream_mode_details(dict[str, Any], Optional):
            Indicates the capacity mode of the data stream. Currently, in Kinesis Data Streams, you can choose between
            an on-demand capacity mode and a provisioned capacity mode for your data streams. Defaults to None.

                * StreamMode (str):
                    Specifies the capacity mode to which you want to set your data stream. Currently, in KinesisData
                    Streams, you can choose between an on-demand capacity mode and a provisioned capacity mode for your
                    data streams.

        shard_level_metrics(list, Optional):
            List of shard-level metrics to disable/ enable. The following are the valid shard-level metrics.
            The value "ALL" enables every metric.
                * IncomingBytes
                * IncomingRecords
                * OutgoingBytes
                * OutgoingRecords
                * WriteProvisionedThroughputExceeded
                * ReadProvisionedThroughputExceeded
                * IteratorAgeMilliseconds
                * ALL

        retention_period_hours(int, Optional):
            IncreaseStreamRetentionPeriod:
                1. Increases the Kinesis data stream's retention period, which is the length of time data records are
                   accessible after they are added to the stream.
                2. If you choose a longer stream retention period, this operation increases the time period during which
                   records that have not yet expired are accessible.
                3. However, it does not make previous, expired data (older than the stream's previous retention period)
                   accessible after the operation has been called.
                4. For example, if a stream's retention period is set to 24 hours and is increased to 168 hours, any
                   data that is older than 24 hours remains inaccessible to consumer applications.

            DecreaseStreamRetentionPeriod:
                1. Decreases the Kinesis data stream's retention period, which is the length of time data records are
                   accessible after they are added to the stream. The minimum value of a stream's retention period is 24
                   hours.

                2. This operation may result in lost data. For example, if the stream's retention period is 48 hours and
                   is decreased to 24 hours, any data already in the stream that is older than 24 hours is inaccessible.

            * The maximum value of a stream's retention period is 8760 hours (365 days).
            * The minimum value of a stream's retention period is 24 hours.

        encryption_type(str, Optional):
            The encryption type to use. The only valid value is KMS .

        key_id (str, Optional):
            The GUID for the customer-managed Amazon Web Services KMS key to use for encryption. This value can
            be a globally unique identifier, a fully specified Amazon Resource Name (ARN) to either an alias or
            a key, or an alias name prefixed by "alias/".

            You can also use a master key owned by Kinesis Data Streams by specifying the alias aws/kinesis.
                * Key ARN example:
                    arn:aws:kms:us-east-1:45678908712456908:key/1818191-7839-9034-3456-45678908712456908
                * Alias ARN example:
                    arn:aws:kms:us-east-1:45678908712456908:alias/MyAliasName
                * Globally unique key ID example:
                    181910-119191-9034-3456-45678908712456908
                * Alias name example:
                    alias/MyAliasName
                * Master key owned by Kinesis Data Streams:
                    alias/aws/kinesis

        stream_arn(str, Optional):
            Specifies the ARN of the data stream whose capacity mode you want to update.

        tags(dict[str, str], Optional):
            The tags to assign to the stream. Defaults to None.
                * Key (str): The key of the tag.
                * Value (str, Optional): The value of the tag.

        timeout(dict, Optional):
            Timeout configuration for AWS Kinesis Stream.

            * create (*dict, Optional*):
                Timeout configuration when creating an AWS Kinesis Stream.

                * delay (*int, Optional*):
                    The amount of time in seconds to wait between attempts. Default value is ``10``.
                * max_attempts (*int, Optional*):
                    Max attempts of waiting for change. Default value is ``18``.

            * update (*dict, Optional*):
                Timeout configuration when updating an AWS Kinesis Stream.

                * delay (*int, Optional*):
                    The amount of time in seconds to wait between attempts. Default value is ``10``.
                * max_attempts (*int, Optional*):
                    Max attempts of waiting for change. Default value is ``18``.


    Request Syntax:
        .. code-block:: sls

            [stream-name]:
              aws.kinesis.stream.present:
                - name: "string"
                - shard_count: "int"
                - resource_id: "string"
                - stream_mode_details: "list"
                - shard_level_metrics: "list"
                - retention_period_hours: "int"
                - encryption_type: "string"
                - key_id: "string"
                - tags:
                    - 'string': "string"
                - timeout:
                    create:
                      delay: int
                      max_attemps: int
                    update:
                      delay: int
                      max_attemps: int


    Returns:
      Dict[str, Any]

    Examples:

        .. code-block:: sls

            stream_is_present:
              aws.kinesis.stream.present:
                - name: 'my-stream'
                - resource_id: 'arn:aws:kinesis:us-west-2:740563969943:stream/test-stream'
                - shard_count: 123
                - stream_mode_details:
                    StreamMode: PROVISIONED
                - shard_level_metrics:
                    - IncomingBytes
                    - IncomingRecords
                    - OutgoingBytes
                    - OutgoingRecords
                    - WriteProvisionedThroughputExceeded
                    - ReadProvisionedThroughputExceeded
                    - IteratorAgeMilliseconds
                    - ALL
                - retention_period_hours: 123
                - encryption_type: 'KMS'
                - key_id: arn:aws:kms:us-east-1:908738389389012:key/234590-7839-9034-3456-9087389389012
                - tags:
                    name: my-stream
                - timeout:
                    create:
                        delay: 10
                        max_attempts: 18
                    update:
                        delay: 10
                        max_attempts: 18
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    plan_state = {}
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.aws.kinesis.stream.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = tuple(before["comment"])
            return result
        result["old_state"] = copy.deepcopy(before["ret"])
        plan_state = copy.deepcopy(result["old_state"])
        input_params = {
            "shard_count": shard_count,
            "shard_level_metrics": shard_level_metrics,
            "stream_mode_details": stream_mode_details,
            "retention_period_hours": retention_period_hours,
            "encryption_type": encryption_type,
            "key_id": key_id,
            "tags": tags,
            "arn": stream_arn,
        }
        update_return = await hub.tool.aws.kinesis.stream.update(
            ctx=ctx,
            name=name,
            current_state=result["old_state"],
            input_map=input_params,
            plan_state=plan_state,
            resource_id=resource_id,
            timeout=timeout,
        )
        result["comment"] = tuple(update_return["comment"])
        if not update_return["result"]:
            result["result"] = False
            return result
        resource_updated = bool(update_return["ret"])
        if resource_updated:
            if ctx.get("test", False):
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.kinesis.stream", name=name
                )
            else:
                result["comment"] += hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.kinesis.stream", name=name
                )

        if tags is not None and tags != result["old_state"].get("tags"):
            # Update tags
            update_tags = await hub.tool.aws.kinesis.tag.update(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=result["old_state"]["tags"],
                new_tags=tags,
            )
            result["comment"] += update_tags["comment"]
            result["result"] = update_tags["result"]
            resource_updated = resource_updated or bool(update_tags["ret"])
            if ctx.get("test", False) and update_tags["ret"]:
                plan_state["tags"] = update_tags["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "shard_count": shard_count,
                    "stream_mode_details": stream_mode_details,
                    "shard_level_metrics": shard_level_metrics,
                    "retention_period_hours": retention_period_hours,
                    "encryption_type": encryption_type,
                    "key_id": key_id,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.kinesis.stream", name=name
            )
            return result
        ret = await hub.exec.boto3.client.kinesis.create_stream(
            ctx,
            StreamName=name,
            ShardCount=shard_count,
            StreamModeDetails=stream_mode_details,
        )
        if not ret["result"] and not ret["ret"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.kinesis.stream", name=name
        )
        resource_id = name

        waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
            ctx, name, timeout, "create"
        )
        if not waiter_ret["result"]:
            result["result"] = False
            result["comment"] += (waiter_ret["comment"],)
            return result

        if tags:
            ret = await hub.exec.boto3.client.kinesis.add_tags_to_stream(
                ctx,
                StreamName=resource_id,
                Tags=tags,
            )
            if not ret["result"] and not ret["ret"]:
                result["comment"] += (ret["comment"],)
                result["result"] = False
                return result
            result["comment"] += ("Added tags to Stream.",)

        if encryption_type is not None and encryption_type == ENCRYPTION_KMS_TYPE:
            if key_id:
                ret = await hub.exec.boto3.client.kinesis.start_stream_encryption(
                    ctx,
                    StreamName=resource_id,
                    EncryptionType=encryption_type,
                    KeyId=key_id,
                )
                if not ret["result"] and not ret["ret"]:
                    result["comment"] += ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] += ("Stream encryption started.",)
                waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                    ctx, name, timeout, "create"
                )
                if not waiter_ret["result"]:
                    result["result"] = False
                    result["comment"] += waiter_ret["comment"]
                    return result

        if shard_level_metrics:
            enable_metrics = (
                await hub.exec.boto3.client.kinesis.enable_enhanced_monitoring(
                    ctx,
                    StreamName=resource_id,
                    ShardLevelMetrics=shard_level_metrics,
                )
            )
            if not enable_metrics["result"]:
                result["comment"] += enable_metrics["comment"]
                result["result"] = False
                return result
            result["comment"] += ("Enabled Enhanced Monitoring for ShardLevelMetrics.",)
            waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                ctx, name, timeout, "create"
            )
            if not waiter_ret["result"]:
                result["result"] = False
                result["comment"] += waiter_ret["comment"]
                return result

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif (not result["old_state"]) or resource_updated:
        after = await hub.exec.aws.kinesis.stream.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not after["result"] or not after["ret"]:
            result["result"] = False
            result["comment"] += after["comment"]
            return result
        result["new_state"] = copy.deepcopy(after["ret"])
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    enforce_consumer_deletion: bool = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=10)),
                        ("max_attempts", int, field(default=18)),
                    ],
                ),
                field(default=None),
            )
        ],
    ) = None,
) -> Dict[str, Any]:
    """Deletes a Kinesis data stream and all its shards and data. You must shut down any applications that are
    operating on the stream before you delete the stream.

    If an application attempts to operate on a deleted stream, it receives the exception ResourceNotFoundException. If
    the stream is in the ACTIVE state, you can delete it. After a DeleteStream request, the specified stream is in the
    DELETING state until Kinesis Data Streams completes the deletion.

    Note: Kinesis Data Streams might continue to accept data read and write operations, such as PutRecord, PutRecords,
    and GetRecords, on a stream in the DELETING state until the stream deletion is complete.

    When you delete a stream, any shards in that stream are also deleted, and any tags are dissociated from the stream.
    You can use the DescribeStreamSummary operation to check the state of the stream, which is returned in StreamStatus.
    DeleteStream has a limit of five transactions per second and per account.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str):
            An identifier of the resource in the provider.

        enforce_consumer_deletion(bool, Optional):
            If this parameter is unset (null) or if you set it to false, and the stream has registered consumers, the
            call to DeleteStream fails with a ResourceInUseException. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for AWS Kinesis Stream.

            * delete (*dict, Optional*):
                Timeout configuration when deleting an AWS Kinesis Stream.

                * delay (*int, Optional*):
                    The amount of time in seconds to wait between attempts. Default value is ``10``.
                * max_attempts (*int, Optional*):
                    Max attempts of waiting for change. Default value is ``18``.

    Request Syntax:
      .. code-block:: sls
           [stream-name]:
              aws.kinesis.stream.absent:
                - name: "string"
                - resource_id: "string"
                - enforce_consumer_deletion: "boolean"
                - timeout:
                    delete:
                      delay: int
                      max_attemps: int

    Returns:
      Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.kinesis.stream.absent:
                - name: value
                - resource_id: value
                - enforce_consumer_deletion: TRUE|FALSE
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.kinesis.stream", name=name
        )
        return result
    before = await hub.exec.aws.kinesis.stream.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.kinesis.stream", name=name
        )
    else:
        result["old_state"] = before["ret"]
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.kinesis.stream", name=name
            )
            return result
        else:
            ret = await hub.exec.boto3.client.kinesis.delete_stream(
                ctx,
                StreamName=resource_id,
                EnforceConsumerDeletion=enforce_consumer_deletion,
            )
            if not ret["result"]:
                result["comment"] = before["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.kinesis.stream", name=name
            )
            waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                ctx,
                name,
                timeout,
                "delete",
                waiter_type="stream_not_exists",
            )
            if not waiter_ret["result"]:
                result["result"] = False
                result["comment"] += waiter_ret["comment"]
                return result
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Pass required params to get a Kinesis stream resource.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.kinesis.stream
    """
    result = {}
    streams = await hub.exec.boto3.client.kinesis.list_streams(ctx)
    if not streams["result"]:
        hub.log.debug(f"Could not describe aws.kinesis.stream(s): {streams['comment']}")
        return result

    if streams["ret"].get("StreamNames"):
        for stream in streams["ret"].get("StreamNames"):
            stream_summary = await hub.exec.aws.kinesis.stream.get(
                ctx=ctx, name=stream, resource_id=stream
            )
            if not stream_summary["result"]:
                hub.log.debug(
                    f"Could not describe aws.kinesis.stream '{stream}: {stream_summary['comment']}"
                )
                continue
            resource_converted = stream_summary["ret"]
            result[resource_converted["resource_id"]] = {
                "aws.kinesis.stream.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_converted.items()
                ]
            }
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


async def search_raw(
    hub,
    ctx,
    name: str,
) -> Dict:
    """
    Provides a summarized description of the specified Kinesis data stream without the shard list. The information
    returned includes the stream name, Amazon Resource Name (ARN), status, record retention period, approximate
    creation time, monitoring, encryption details, and open shard count.

    Args:
        name(str):
            AWS Kinesis Stream name to uniquely identify the resource.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.kinesis.describe_stream_summary(
        ctx,
        StreamName=name,
    )
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


async def update(
    hub,
    ctx,
    name: str,
    current_state: Dict[str, Any],
    input_map: Dict[str, Any],
    resource_id: str,
    plan_state: Dict[str, Any],
    timeout: Dict[str, Any],
):
    r"""
    If the input supplied retention is less than the existing, here we decerase the Kinesis data stream's retention
    period, which is the length of time data records are accessible after they are added to the stream. The minimum
    value of a stream's retention period is 24 hours. This operation may result in lost data. For example, if the
    stream's retention period is 48 hours and is decreased to 24 hours, any data already in the stream that is older
    than 24 hours is inaccessible. Conversely, if the input retention is more than the existing, then we increase
    Kinesis data stream's retention period, which is the length of time data records are accessible after they are
    added to the stream. The maximum value of a stream's retention period is 8760 hours (365 days). If you choose a
    longer stream retention period, this operation increases the time period during which records that have not yet
    expired are accessible. However, it does not make previous, expired data (older than the stream's previous retention
     period) accessible after the operation has been called. For example, if a stream's retention period is set to 24
    hours and is increased to 168 hours, any data that is older than 24 hours remains inaccessible to consumer
    applications. Checks to see if enhanced monitoring needs to be enabled/ disabled for certain ShardLevelMetrics.
    Enables or updates server-side encryption using an Amazon Web Services KMS key for a specified stream. Starting/
    stopping encryption is an asynchronous operation. Upon receiving the request, Kinesis Data Streams returns
    immediately and sets the status of the stream to UPDATING . After the update is complete, Kinesis Data Streams sets
    the status of the stream back to ACTIVE . Updating or applying encryption normally takes a few seconds to complete,
    but it can take minutes. You can continue to read and write data to your stream while its status is UPDATING . Once
    status of the stream is ACTIVE , encryption begins for records written to the stream. API Limits: You can
    successfully apply a new Amazon Web Services KMS key for server-side encryption 25 times in a rolling 24-hour period.
    Note: It can take up to 5 seconds after the stream is in an ACTIVE status before all records written to the stream
    are encrypted. After you enable encryption, you can verify that encryption is applied by inspecting the API response
     from PutRecord or PutRecords. Disables server-side encryption for a specified stream. Updates the shard count of
    the specified stream to the specified number of shards.
    This operation has the following default limits. By default, you cannot do the following:
        a. Scale more than ten times per rolling 24-hour period per stream
        b. Scale up to more than double your current shard count for a stream
        c. Scale down below half your current shard count for a stream
        d. Scale up to more than 10000 shards in a stream
        e. Scale a stream with more than 10000 shards down unless the result is less than 10000 shards
        f. Scale up to more than the shard limit for your account

    Updates the capacity mode of the data stream. Currently, in Kinesis Data Streams, you can choose between an
    on-demand capacity mode and a provisioned capacity mode for your data stream.

    Args:
        name(str):
            The name of the AWS ElasticLoadBalancingv2 Listener.

        current_state(dict[str, Any]):
            response returned by describe on an AWS ElasticLoadBalancingv2 Listener

        input_map(dict[str, Any]):
            a dictionary with newly passed values of params.

        resource_id(str):
            AWS ElasticLoadBalancingv2 Listener Amazon Resource Name (ARN).

        plan_state(dict[str, Any]):
            idem --test state for update on AWS ElasticLoadBalancingv2 Listener.

        timeout(dict, optional): Timeout configuration for waiting for Aws Certificate to get issued.
            * describe (Dict) -- Timeout configuration for describing certificate
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.
    Returns:
        {"result": True|False, "comment": A message List, "ret": None}
    """
    result = dict(comment=[], result=True, ret=[])
    if not ctx.get("test", False):
        if input_map.get("shard_count"):
            if current_state.get("shard_count") and (
                input_map.get("shard_count") != current_state.get("shard_count")
            ):
                update_shard_count = (
                    await hub.exec.boto3.client.kinesis.update_shard_count(
                        ctx,
                        StreamName=resource_id,
                        TargetShardCount=input_map.get("shard_count"),
                        ScalingType="UNIFORM_SCALING",
                    )
                )
                if not update_shard_count["result"] and not update_shard_count["ret"]:
                    result["comment"] = list(update_shard_count["comment"])
                    result["result"] = False
                    return result
                result["comment"].append("Updated Shard Count.")
                result["ret"].append(
                    {
                        "update_shard_count": update_shard_count["ret"].get(
                            "TargetShardCount"
                        )
                    }
                )
                waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                    ctx, name, timeout, "update"
                )
                if not waiter_ret["result"]:
                    result["result"] = False
                    result["comment"] += waiter_ret["comment"]
                    return result

        if input_map.get("stream_mode_details"):
            if current_state.get("stream_mode_details") and (
                input_map.get("stream_mode_details")
                != current_state.get("stream_mode_details")
            ):
                update_stream_mode = (
                    await hub.exec.boto3.client.kinesis.update_stream_mode(
                        ctx,
                        StreamARN=input_map.get("arn"),
                        StreamModeDetails=input_map.get("stream_mode_details"),
                    )
                )
                if not update_stream_mode["result"] and not update_stream_mode["ret"]:
                    result["comment"] = list(update_stream_mode["comment"])
                    result["result"] = False
                    return result
                result["comment"].append("Updated Stream Mode.")
                result["ret"].append(
                    {"update_stream_mode": input_map.get("stream_mode_details")}
                )
                waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                    ctx, name, timeout, "update"
                )
                if not waiter_ret["result"]:
                    result["result"] = False
                    result["comment"] += waiter_ret["comment"]
                    return result

        if input_map.get("retention_period_hours"):
            current_retention_period = current_state.get("retention_period_hours")
            if current_retention_period != input_map.get("retention_period_hours"):
                if current_retention_period < input_map.get("retention_period_hours"):
                    increase_period = await hub.exec.boto3.client.kinesis.increase_stream_retention_period(
                        ctx,
                        StreamName=resource_id,
                        RetentionPeriodHours=input_map.get("retention_period_hours"),
                    )
                    if not increase_period["result"] and not increase_period["ret"]:
                        result["comment"] = list(increase_period["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append("Increased Retention Period.")
                    result["ret"].append(
                        {
                            "increase_stream_retention_period": input_map.get(
                                "retention_period_hours"
                            )
                        }
                    )
                    waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                        ctx, name, timeout, "update"
                    )
                    if not waiter_ret["result"]:
                        result["result"] = False
                        result["comment"] += waiter_ret["comment"]
                        return result

                else:
                    decrease_period = await hub.exec.boto3.client.kinesis.decrease_stream_retention_period(
                        ctx,
                        StreamName=resource_id,
                        RetentionPeriodHours=input_map.get("retention_period_hours"),
                    )
                    if not decrease_period["result"] and not decrease_period["ret"]:
                        result["comment"] = list(decrease_period["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append("Decreased Retention Period.")
                    result["ret"].append(
                        {
                            "decrease_stream_retention_period": input_map.get(
                                "retention_period_hours"
                            )
                        }
                    )
                    waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                        ctx, name, timeout, "update"
                    )
                    if not waiter_ret["result"]:
                        result["result"] = False
                        result["comment"] += waiter_ret["comment"]
                        return result

        if input_map.get("shard_level_metrics"):
            if not current_state.get("shard_level_metrics") or set(
                current_state.get("shard_level_metrics")
            ) != set(input_map.get("shard_level_metrics")):
                to_disable = list(
                    set(current_state.get("shard_level_metrics", []))
                    - set(input_map.get("shard_level_metrics"))
                )
                to_enable = list(
                    set(input_map.get("shard_level_metrics"))
                    - set(current_state.get("shard_level_metrics", []))
                )

                if to_disable:
                    disable_metrics = (
                        await hub.exec.boto3.client.kinesis.disable_enhanced_monitoring(
                            ctx,
                            StreamName=resource_id,
                            ShardLevelMetrics=to_disable,
                        )
                    )
                    if not disable_metrics["result"]:
                        result["comment"] += list(disable_metrics["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append(
                        "Disabled Enhanced Monitoring for ShardLevelMetrics."
                    )
                    result["ret"].append(
                        {
                            "disable_metrics": disable_metrics["ret"][
                                "DesiredShardLevelMetrics"
                            ]
                        }
                    )
                    waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                        ctx, name, timeout, "update"
                    )
                    if not waiter_ret["result"]:
                        result["result"] = False
                        result["comment"] += waiter_ret["comment"]
                        return result

                if to_enable:
                    enable_metrics = (
                        await hub.exec.boto3.client.kinesis.enable_enhanced_monitoring(
                            ctx,
                            StreamName=resource_id,
                            ShardLevelMetrics=to_enable,
                        )
                    )
                    if not enable_metrics["result"]:
                        result["comment"] += list(enable_metrics["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append(
                        "Enabled Enhanced Monitoring for ShardLevelMetrics."
                    )
                    result["ret"].append(
                        {
                            "enable_metrics": enable_metrics["ret"][
                                "DesiredShardLevelMetrics"
                            ]
                        }
                    )
                    waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                        ctx, name, timeout, "update"
                    )
                    if not waiter_ret["result"]:
                        result["result"] = False
                        result["comment"] += waiter_ret["comment"]
                        return result

        if input_map.get("encryption_type"):
            if input_map.get("encryption_type") == "NONE":
                if current_state.get("encryption_type") == "KMS":
                    stop_encryption = (
                        await hub.exec.boto3.client.kinesis.stop_stream_encryption(
                            ctx,
                            StreamName=resource_id,
                            EncryptionType="KMS",
                            KeyId=input_map.get("key_id"),
                        )
                    )
                    if not stop_encryption["result"]:
                        result["comment"] += list(stop_encryption["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append("Stopped Stream Encryption.")
                    result["ret"].append(
                        {"stop_stream_encryption": input_map.get("encryption_type")}
                    )
                    waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                        ctx, name, timeout, "update"
                    )
                    if not waiter_ret["result"]:
                        result["result"] = False
                        result["comment"] += waiter_ret["comment"]
                        return result

            elif input_map.get("encryption_type") == "KMS":
                if current_state.get("encryption_type") == "NONE":
                    start_encryption = (
                        await hub.exec.boto3.client.kinesis.start_stream_encryption(
                            ctx,
                            StreamName=resource_id,
                            EncryptionType="KMS",
                            KeyId=input_map.get("key_id"),
                        )
                    )
                    if not start_encryption["result"]:
                        result["comment"] += list(start_encryption["comment"])
                        result["result"] = False
                        return result
                    result["comment"].append("Started Stream Encryption.")
                    result["ret"].append(
                        {"start_stream_encryption": input_map.get("encryption_type")}
                    )
                    waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                        ctx, name, timeout, "update"
                    )
                    if not waiter_ret["result"]:
                        result["result"] = False
                        result["comment"] += waiter_ret["comment"]
                        return result

                else:
                    if current_state.get("key_id") != input_map.get("key_id"):
                        start_encryption = (
                            await hub.exec.boto3.client.kinesis.start_stream_encryption(
                                ctx,
                                StreamName=resource_id,
                                EncryptionType="KMS",
                                KeyId=input_map.get("key_id"),
                            )
                        )
                        if not start_encryption["result"]:
                            result["comment"] += list(start_encryption["comment"])
                            result["result"] = False
                            return result
                        result["comment"].append("Started Stream Encryption.")
                        result["ret"].append(
                            {
                                "start_stream_encryption": input_map.get(
                                    "encryption_type"
                                )
                            }
                        )
                        waiter_ret = await hub.tool.aws.kinesis.stream.kinesis_waiter(
                            ctx, name, timeout, "update"
                        )
                        if not waiter_ret["result"]:
                            result["result"] = False
                            result["comment"] += waiter_ret["comment"]
                            return result
    else:
        update_params = OrderedDict(
            {
                "name": name,
                "shard_count": input_map.get("shard_count"),
                "stream_mode_details": input_map.get("stream_mode_details"),
                "shard_level_metrics": input_map.get("shard_level_metrics"),
                "retention_period_hours": input_map.get("retention_period_hours"),
                "encryption_type": input_map.get("encryption_type"),
                "key_id": input_map.get("key_id"),
                "resource_id": resource_id,
                "tags": input_map.get("tags"),
            }
        )
        for key, value in update_params.items():
            if value is not None:
                plan_state[key] = value
        result["ret"] = plan_state
    return result


async def kinesis_waiter(
    hub,
    ctx,
    name: str,
    timeout: Dict,
    operation_type: str,
    waiter_type: str = None,
):
    """

    Waiter to wait for the Kinesis Stream to become active.

        Args:
           name(Text): An Idem name of the resource.
           timeout(Dict, optional): Timeout configuration for creating or updating Kinesis Stream.
            * create (Dict) -- Timeout configuration for creating Kinesis Stream
                * delay(int, default=10) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=18) -- Customized timeout configuration containing delay and max attempts.
            * update (str) -- Timeout configuration for updating Kinesis Stream
                * delay(int, default=10) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=18) -- Customized timeout configuration containing delay and max attempts.
            * operation_type(str): create or update operation
            * waiter_type(str): StreamNotExists or StreamExists

        Returns:
            Dict[str, Any]
    """

    result = dict(comment=[], result=True, ret={})
    if not waiter_type:
        waiter_type = "stream_exists"
    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=10,
        default_max_attempts=18,
        timeout_config=timeout.get(operation_type) if timeout else None,
    )
    hub.log.debug(f"Waiting on aws.kinesis.stream '{name}' to become active")
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "kinesis",
            waiter_type,
            StreamName=name,
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] += list(str(e))
        result["result"] = False
    return result



"""Exec functions for AWS API Gateway v2 Integration resources."""
from typing import Any
from typing import Dict


async def get(
    hub, ctx, resource_id: str, api_id: str, name: str = None
) -> Dict[str, Any]:
    """Gets an Integration.

    Args:
        resource_id(str):
            The integration ID.

        api_id(str):
            The API identifier.

        name(str, Optional):
            The name of the Idem state.

    Returns:
        Dict[str, Any]:
            Returns an integration.

    Examples:
        Call from the CLI:

        .. code-block:: bash

            idem exec aws.apigatewayv2.integration.get resource_id="integration_id" api_id="api_id" name="my_resource"

        Call from code:

        .. code-block:: python

            async def my_func(hub, ctx, resource_id:str, api_id: str, name:str = None):
                await hub.exec.aws.apigatewayv2.integration.get(ctx, resource_id, api_id, name)

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.apigatewayv2.integration.get
                - kwargs:
                    resource_id: integration_id
                    api_id: api_id
                    name: my_resource
    """
    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.apigatewayv2.get_integration(
        ctx, ApiId=api_id, IntegrationId=resource_id
    )
    if not ret["result"]:
        if "NotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.apigatewayv2.integration", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result
    result[
        "ret"
    ] = hub.tool.aws.apigatewayv2.integration.convert_raw_integration_to_present(
        api_id=api_id,
        raw_resource=ret["ret"],
        idem_resource_name=name if name is not None else resource_id,
    )
    return result



from typing import Any
from typing import Dict


def convert_raw_event_destination_to_present(
    hub, configuration_set_name: str, name: str, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Util functions to convert raw resource state to present input format for event destination.
    """
    ret = {
        "name": name,
        "configuration_set_name": configuration_set_name,
        "resource_id": name,
    }
    if "Name" in raw_resource:
        raw_resource.pop("Name")
    ret["event_destination"] = raw_resource.copy()
    return ret



"""
Util functions to convert raw resource state from AWS Cloudformation Stack to present input format.
"""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_stack_to_present(
    hub,
    raw_resource: Dict[str, Any],
    resource_id: str,
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    resource_parameters = OrderedDict(
        {
            "TemplateBody": "template_body",
            "TemplateUrl": "template_url",
            "Parameters": "parameters",
            "DisableRollback": "disable_rollback",
            "RollbackConfiguration": "rollback_configuration",
            "TimeoutInMinutes": "timeout",
            "NotificationARNs": "notification_arns",
            "Capabilities": "capabilities",
            "ResourceTypes": "resource_types",
            "RoleARN": "role_arn",
            "OnFailure": "on_failure",
            "StackPolicyBody": "stack_policy_body",
            "StackPolicyURL": "stack_policy_url",
            "Tags": "tags",
            "ClientRequestToken": "client_request_token",
            "EnableTerminationProtection": "enable_termination_protection",
            "StackStatus": "stack_status",
            "StackDriftStatus": "stack_drift_status",
            "RetainResources": "retain_resources",
        }
    )

    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated


async def update(
    hub,
    ctx,
    stack_name: str,
    old_state: Dict[str, Any],
    template_body: str = None,
    template_url: str = None,
    parameters: List = None,
    rollback_configuration: Dict = None,
    resource_types: List = None,
    notification_arns: List = None,
    capabilities: List = None,
    role_arn: str = None,
    stack_policy_body: str = None,
    stack_policy_url: str = None,
    tags: List = None,
    client_request_token: str = None,
    disable_rollback: bool = None,
    timeout: dict = None,
) -> Dict[str, any]:
    """
    Args:
        stack_name(str):
            A name, ID to identify the resource.

        old_state(dict):
            Previous state of the resource.

        template_body(str, Optional):
            Structure containing the template body with a minimum length of 1 byte and a maximum length of 51,200 bytes.
            For more information, go to Template anatomy in the CloudFormation User Guide. Conditional: You must specify
            either the TemplateBody or the TemplateURL parameter, but not both. Defaults to None.

        template_url(str, Optional):
            Location of file containing the template body. The URL must point to a template (max size: 460,800 bytes)
            that's located in an Amazon S3 bucket or a Systems Manager document. For more information, go to the
            Template anatomy in the CloudFormation User Guide. Conditional: You must specify either the TemplateBody or
            the TemplateURL parameter, but not both. Defaults to None.

        parameters(list, Optional):
            A list of Parameter structures that specify input parameters for the stack. For more information, see the
            Parameter data type. Defaults to None.

        capabilities(list, Optional):
            In some cases, you must explicitly acknowledge that your stack template contains certain capabilities in order
            for CloudFormation to create the stack. CAPABILITY_IAM and CAPABILITY_NAMED_IAM  Some stack templates might
            include resources that can affect permissions in your Amazon Web Services account; for example, by creating
            new Identity and Access Management (IAM) users. For those stacks, you must explicitly acknowledge this by
            specifying one of these capabilities. The following IAM resources require you to specify either the CAPABILITY_IAM
            or CAPABILITY_NAMED_IAM capability. If you have IAM resources, you can specify either capability. If you have
            IAM resources with custom names, you must specify CAPABILITY_NAMED_IAM. If you don't specify either of these
            capabilities, CloudFormation returns an InsufficientCapabilities error. If your stack template contains these
            resources, we recommend that you review all permissions associated with them and edit their permissions if
            necessary. AWS::IAM::AccessKey, AWS::IAM::Group, AWS::IAM::InstanceProfile AWS::IAM::Policy, AWS::IAM::Role,
            AWS::IAM::User, AWS::IAM::UserToGroupAddition. For more information, see Acknowledging IAM Resources in
            CloudFormation Templates. CAPABILITY_AUTO_EXPAND: Some templates contain macros. Macros perform custom processing
            on templates; this can include simple actions like find-and-replace operations, all the way to extensive
            transformations of entire templates. Because of this, users typically create a change set from the processed
            template, so that they can review the changes resulting from the macros before actually creating the stack.
            If your stack template contains one or more macros, and you choose to create a stack directly from the processed
            template, without first reviewing the resulting changes in a change set, you must acknowledge this capability.
            This includes the AWS::Include and AWS::Serverless transforms, which are macros hosted by CloudFormation.
            If you want to create a stack from a stack template that contains macros and nested stacks, you must create
            the stack directly from the template using this capability. You should only create stacks directly from a stack
            template that contains macros if you know what processing the macro performs. Each macro relies on an underlying
            Lambda service function for processing stack templates. Be aware that the Lambda function owner can update
            the function operation without CloudFormation being notified. For more information, see Using CloudFormation
            macros to perform custom processing on templates. Defaults to None.

        resource_types(list, Optional):
            The template resource types that you have permissions to work with for this create stack action, such as
            AWS::EC2::Instance, AWS::EC2::*, or Custom::MyCustomInstance. Use the following syntax to describe template
            resource types: AWS::* (for all Amazon Web Services resources), Custom::* (for all custom resources),
            Custom::logical_ID  (for a specific custom resource), AWS::service_name::* (for all resources of a particular
            Amazon Web Services service), and AWS::service_name::resource_logical_ID  (for a specific Amazon Web Services
            resource). If the list of resource types doesn't include a resource that you're creating, the stack creation
            fails. By default, CloudFormation grants permissions to all resource types. Identity and Access Management
            (IAM) uses this parameter for CloudFormation-specific condition keys in IAM policies. For more information,
            see Controlling Access with Identity and Access Management. Defaults to None.

        role_arn(str, Optional):
            The Amazon Resource Name (ARN) of an Identity and Access Management (IAM) role that CloudFormation assumes
            to create the stack. CloudFormation uses the role's credentials to make calls on your behalf. CloudFormation
            always uses this role for all future operations on the stack. Provided that users have permission to operate
            on the stack, CloudFormation uses this role even if the users don't have permission to pass it. Ensure that
            the role grants least privilege. If you don't specify a value, CloudFormation uses the role that was
            previously associated with the stack. If no role is available, CloudFormation uses a temporary session that's
            generated from your user credentials. Defaults to None.

        rollback_configuration(dict, Optional):
            The rollback triggers for CloudFormation to monitor during stack creation and updating operations, and for
            the specified monitoring period afterwards. Defaults to None.

        stack_policy_body(str, Optional):
            Structure containing the stack policy body. For more information, go to  Prevent Updates to Stack Resources
            in the CloudFormation User Guide. You can specify either the StackPolicyBody or the StackPolicyURL parameter,
            but not both. Defaults to None.

        stack_policy_url(str, Optional):
            Location of a file containing the stack policy. The URL must point to a policy (maximum size: 16 KB) located
            in an S3 bucket in the same Region as the stack. You can specify either the StackPolicyBody or the StackPolicyURL
            parameter, but not both. Defaults to None.

        notification_arns(list, Optional):
            The Amazon Simple Notification Service (Amazon SNS) topic ARNs to publish stack related events. You can find
            your Amazon SNS topic ARNs using the Amazon SNS console or your Command Line Interface (CLI). Defaults to
            None.

        tags(list, Optional):
            Key-value pairs to associate with this stack. CloudFormation also propagates these tags to the resources
            created in the stack. A maximum number of 50 tags can be specified. Defaults to None.

        disable_rollback(bool, Optional):
            Set to true to disable rollback of the stack if stack creation failed. You can specify either DisableRollback
            or OnFailure, but not both. Default: false. Defaults to None.

        client_request_token(str, Optional):
            A unique identifier for this CreateStack request. Specify this token if you plan to retry requests so that
            CloudFormation knows that you're not attempting to create a stack with the same name. You might retry
            CreateStack requests to ensure that CloudFormation successfully received them. All events initiated by a
            given stack operation are assigned the same client request token, which you can use to track operations.
            For example, if you execute a CreateStack operation with the token token1, then all the StackEvents generated
            by that operation will have ClientRequestToken set as token1. In the console, stack operations display the
            client request token on the Events tab. Stack operations that are initiated from the console use the token
            format Console-StackOperation-ID, which helps you easily identify the stack operation. For example, if you
            create a stack using the console, each stack event would be assigned the same token in the following format:
            Console-CreateStack-7f59c3cf-00d2-40c7-b2ff-e75db0987002. Defaults to None.

        timeout(dict, Optional):
            * update(dict, Optional):
                Timeout configuration for updating AWS Cloudformation Stack.

                * delay (int, Optional): The amount of time in seconds to wait between attempts.

                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.

        Returns:
            {"result": True|False, "comment": "A message tuple", "ret": None}
    """

    result = dict(comment=(), result=True, ret=None)
    new_state = copy.deepcopy(old_state)

    parameters, new_state["parameters"] = (
        (parameters, parameters)
        if not (
            hub.tool.aws.state_comparison_utils.are_lists_identical(
                parameters, old_state.get("parameters")
            )
        )
        else (old_state.get("parameters"), old_state.get("parameters"))
    )
    disable_rollback, new_state["disable_rollback"] = (
        (disable_rollback, disable_rollback)
        if disable_rollback != old_state.get("disable_rollback")
        else (old_state.get("disable_rollback"), old_state.get("disable_rollback"))
    )
    template_url, new_state["template_url"] = (
        (template_url, template_url)
        if template_url != old_state.get("template_url")
        else (old_state.get("template_url"), old_state.get("template_url"))
    )
    template_body, new_state["template_body"] = (
        (template_body, template_body)
        if template_body != old_state.get("template_body")
        else (old_state.get("template_body"), old_state.get("template_body"))
    )
    tags, new_state["tags"] = (
        (tags, tags)
        if tags != old_state.get("tags")
        else (old_state.get("tags"), old_state.get("tags"))
    )
    capabilities, new_state["capabilities"] = (
        (capabilities, capabilities)
        if capabilities != old_state.get("capabilities")
        else (old_state.get("capabilities"), old_state.get("capabilities"))
    )
    rollback_configuration, new_state["rollback_configuration"] = (
        (rollback_configuration, rollback_configuration)
        if rollback_configuration != old_state.get("rollback_configuration")
        else (
            old_state.get("rollback_configuration"),
            old_state.get("rollback_configuration"),
        )
    )
    role_arn, new_state["role_arn"] = (
        (role_arn, role_arn)
        if role_arn != old_state.get("role_arn")
        else (old_state.get("role_arn"), old_state.get("role_arn"))
    )
    notification_arns, new_state["notification_arns"] = (
        (notification_arns, notification_arns)
        if notification_arns != old_state.get("notification_arns")
        else (old_state.get("notification_arns"), old_state.get("notification_arns"))
    )
    stack_policy_body, new_state["stack_policy_body"] = (
        (stack_policy_body, stack_policy_body)
        if stack_policy_body != old_state.get("stack_policy_body")
        else (old_state.get("stack_policy_body"), old_state.get("stack_policy_body"))
    )
    stack_policy_url, new_state["stack_policy_url"] = (
        (stack_policy_url, stack_policy_url)
        if stack_policy_url != old_state.get("stack_policy_url")
        else (old_state.get("stack_policy_url"), old_state.get("stack_policy_url"))
    )
    resource_types, new_state["resource_types"] = (
        (resource_types, resource_types)
        if resource_types != old_state.get("resource_types")
        else (old_state.get("resource_types"), old_state.get("resource_types"))
    )
    client_request_token, new_state["client_request_token"] = (
        (client_request_token, client_request_token)
        if client_request_token != old_state.get("client_request_token")
        else (
            old_state.get("client_request_token"),
            old_state.get("client_request_token"),
        )
    )

    if ctx.get("test", False):
        if new_state != old_state:
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.cloudformation.stack", name=old_state["name"]
            )
            result["ret"] = new_state
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.cloudformation.stack", name=old_state["name"]
            )
        return result

    elif new_state != old_state:
        ret = await hub.exec.boto3.client.cloudformation.update_stack(
            ctx,
            StackName=stack_name,
            TemplateBody=template_body,
            TemplateURL=template_url,
            Parameters=parameters,
            StackPolicyBody=stack_policy_body,
            StackPolicyURL=stack_policy_url,
            Capabilities=capabilities,
            ResourceTypes=resource_types,
            RoleARN=role_arn,
            RollbackConfiguration=rollback_configuration,
            Tags=tags,
            ClientRequestToken=client_request_token,
            DisableRollback=disable_rollback,
        )
        if not ret["result"]:
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
        result["comment"] = hub.tool.aws.comment_utils.update_comment(
            resource_type="aws.cloudformation.stack", name=stack_name
        )
        update_waiter_acceptors = [
            {
                "matcher": "pathAll",
                "expected": "UPDATE_COMPLETE",
                "state": "success",
                "argument": "Stacks[].StackStatus",
            },
            {
                "matcher": "pathAll",
                "expected": "UPDATE_IN_PROGRESS",
                "state": "retry",
                "argument": "Stacks[].StackStatus",
            },
            {
                "matcher": "pathAll",
                "expected": "ROLLBACK_COMPLETE",
                "state": "retry",
                "argument": "Stacks[].StackStatus",
            },
            {
                "matcher": "pathAll",
                "expected": "CREATE_COMPLETE",
                "state": "retry",
                "argument": "Stacks[].StackStatus",
            },
            {
                "matcher": "pathAll",
                "expected": "CREATE_IN_PROGRESS",
                "state": "retry",
                "argument": "Stacks[].StackStatus",
            },
            {
                "matcher": "pathAll",
                "expected": "ROLLBACK_COMPLETE",
                "state": "retry",
                "argument": "Stacks[].StackStatus",
            },
        ]
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=10,
            default_max_attempts=60,
            timeout_config=timeout.get("update") if timeout else None,
        )
        endpoint_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="StackUpdateComplete",
            operation="DescribeStacks",
            argument=["Stacks[].StackStatus"],
            acceptors=update_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "cloudformation"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "cloudformation",
                "StackUpdateComplete",
                endpoint_waiter,
                StackName=stack_name,
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = str(e)
            result["result"] = False

    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.cloudformation.stack", name=old_state["name"]
        )

    return result



"""Exec module for managing Amazon Cloudformation Stacks."""


async def get(
    hub,
    ctx,
    name: str,
    resource_id: str,
):
    """
    Get a Cloudformation Stack from AWS.

    Args:
        name(str):
            The name of the Cloudformation Stack.

        resource_id(str):
            The StackID of the Cloudformation Stack.

    Examples:

        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec aws.cloudformation.stack.get name="name" resource_id="unmanaged_stack"


        Calling this exec module function from within a state module in pure python

        .. code-block:: yaml

            my_unmanaged_stack:
              exec.run:
                - path: aws.cloudformation.stack.get
                - kwargs:
                    name: test-stack-name
                    resource_id: ax0li
    """

    result = dict(comment=[], ret=None, result=True)
    ret = await hub.exec.boto3.client.cloudformation.describe_stacks(
        ctx, StackName=name
    )

    if (not ret["result"]) or (
        ret["ret"]["Stacks"][0].get("StackStatus") == "DELETE_COMPLETE"
    ):
        if ("ValidationError" in str(ret["comment"])) or (
            ret["ret"]["Stacks"][0].get("StackStatus") == "DELETE_COMPLETE"
        ):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.cloudformation.stack", name=name
                )
            )
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    resource = ret["ret"]["Stacks"][0]
    if len(ret["ret"]["Stacks"]) > 1:
        result["comment"].append(
            hub.tool.aws.comment_utils.find_more_than_one(
                resource_type="aws.cloudformation.stack", resource_id=resource_id
            )
        )

    elif resource.get("StackStatus") != "DELETE_COMPLETE":
        result["ret"] = hub.tool.aws.cloudformation.stack.convert_raw_stack_to_present(
            raw_resource=ret["ret"]["Stacks"][0],
            resource_id=resource_id,
            idem_resource_name=name,
        )

        return result



from typing import Any
from typing import Dict
from typing import List
from typing import Set

ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS = {
    "DestinationCidrBlock",
    "DestinationIpv6CidrBlock",
    "DestinationPrefixListId",
}

ALLOWED_ROUTE_TARGET_GATEWAYS = {
    "VpcEndpointId",
    "EgressOnlyInternetGatewayId",
    "GatewayId",
    "InstanceId",
    "NatGatewayId",
    "TransitGatewayId",
    "LocalGatewayId",
    "CarrierGatewayId",
    "NetworkInterfaceId",
    "RouteTableId",
    "VpcPeeringConnectionId",
    "CoreNetworkArn",
}

ALLOWED_DEFAULT_ROUTE_TARGET_GATEWAYS = {
    "VpcEndpointId",
    "GatewayId",
    "InstanceId",
    "NetworkInterfaceId",
}


async def search_raw(hub, ctx, resource_id: str = None, filters: List = None) -> Dict:
    """
    Fetch one or more route table from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        resource_id(Text, optional): AWS route table ID to identify the resource.
        filters(list, optional): One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_route_tables

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["result"] = False
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    route_table_args = {}
    if boto3_filter:
        route_table_args["Filters"] = boto3_filter

    # Including None as RouteTableIds, throws ParamValidationError for describe_route_tables
    if resource_id:
        route_table_args["RouteTableIds"] = [resource_id]

    ret = await hub.exec.boto3.client.ec2.describe_route_tables(ctx, **route_table_args)
    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result


def get_route_table_routes_modifications(
    hub, old_routes: List[Dict], new_routes: List[Dict]
):
    """
    Given old routes i.e, the routes of current route table and new_routes
    i.e, the routes we want to modify to, this function will compare differences between
    old and new routes and return routes to be added and removed

    Args:
        hub:
        old_routes(List):  routes of existing route table
        new_routes(List): new routes to update

    Returns:
        Dict[str, List]

    """
    result = dict(
        comment=(),
        routes_to_add=[],
        routes_to_delete=[],
        routes_to_replace=[],
        result=True,
    )
    routes_to_add = []
    routes_to_delete = []
    routes_to_replace = []
    old_routes_map = {}
    default_destination_cidr_block: str = ""
    # Creates a map of destination cidr block -> route to easily retrieve further
    for route in old_routes:
        if route.get("State") == "active":
            # Route table route allows three types of destinations
            # 1) DestinationCidrBlock,
            # 2) DestinationIpv6CidrBlock,
            # 3) DestinationPrefixListId,
            # We need to determine which type of  destination is used
            destination_cidr_key = get_common_key_in_dict(
                route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
            )
            old_routes_map[route.get(destination_cidr_key)] = route
            if route.get("Origin") == "CreateRouteTable":
                default_destination_cidr_block = route.get(destination_cidr_key)
    new_destination_cidr_blocks = list()
    for new_route in new_routes:
        # Loop through new routes and determine which destination is used
        destination_cidr_key = get_common_key_in_dict(
            new_route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
        )
        destination_cidr_block = new_route.get(destination_cidr_key)

        # checking for duplicate destination cidr blocks. There can be only unique destination cidr block
        if destination_cidr_block in new_destination_cidr_blocks:
            result["result"] = False
            result["comment"] = (
                "Duplicate destination cidr blocks are not allowed, they should be unique.",
            )
            return result
        else:
            new_destination_cidr_blocks.append(destination_cidr_block)
        allowed_targets = (
            ALLOWED_DEFAULT_ROUTE_TARGET_GATEWAYS
            if destination_cidr_block == default_destination_cidr_block
            else ALLOWED_ROUTE_TARGET_GATEWAYS
        )
        new_target_key = get_common_key_in_dict(new_route, allowed_targets)
        if new_target_key is None or destination_cidr_key is None:
            result["result"] = False
            result["comment"] = (
                "There should be at least one valid destination cidr block and one target",
            )
            return result
        new_target_value = new_route.get(new_target_key)
        # Destination cidr is unique for a route. After retrieving destination cidr check in old routes
        # if the new route cidr block is not present in old routes we associate it to route table
        # if the destination cidr is already present in old routes we need to check if there is any changes in target
        # The allowed targets can be any one of Internet gateway, Virtual private gateway and many more.
        # If the target key is modified or value is modified we need to delete old route and create
        # a new route with updated values
        # example the target might be changed from an Internet gateway to Transit gateway the key changes
        # So we need to delete Internet gateway route and create Transit gateway route
        if destination_cidr_block not in old_routes_map:
            routes_to_add.append(
                {
                    destination_cidr_key: destination_cidr_block,
                    new_target_key: new_target_value,
                }
            )
        else:
            old_route = old_routes_map.get(destination_cidr_block)
            old_target_key = get_common_key_in_dict(
                old_route, ALLOWED_ROUTE_TARGET_GATEWAYS
            )
            old_target_value = old_route.get(old_target_key)
            if old_target_key != new_target_key or old_target_value != new_target_value:
                routes_to_replace.append(
                    {
                        destination_cidr_key: destination_cidr_block,
                        new_target_key: new_target_value,
                    }
                )
            del old_routes_map[destination_cidr_block]

    # delete the remaining routes which are still present in old routes and not in new routes
    for old_route in old_routes_map.values():
        destination_cidr_key = get_common_key_in_dict(
            old_route, ALLOWED_ROUTE_DESTINATION_CIDR_BLOCKS
        )
        destination_cidr_block = old_route.get(destination_cidr_key)
        if old_route.get("Origin") != "CreateRouteTable":
            routes_to_delete.append({destination_cidr_key: destination_cidr_block})

    result["routes_to_add"] = routes_to_add
    result["routes_to_delete"] = routes_to_delete
    result["routes_to_replace"] = routes_to_replace
    return result


def get_common_key_in_dict(dict1: Dict, target_keys_list: Set):
    common_keys = set(dict1.keys()).intersection(target_keys_list)
    return common_keys.pop() if common_keys else None


def get_route_table_association_by_id(
    hub,
    resources: List,
    gateway_id: str = None,
    subnet_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    result = {"result": False, "resource_id": resource_id}
    if resource_id:
        key = "RouteTableAssociationId"
        value = resource_id
    elif subnet_id:
        key = "SubnetId"
        value = subnet_id
    else:
        key = "GatewayId"
        value = gateway_id
    if resources:
        for resource in resources:
            if key in resource and resource.get(key) == value:
                # Do not return ones that are in disassociated or disassociating state
                if resource["AssociationState"]["State"] == "associated":
                    result["ret"] = resource
                    result["result"] = True
                    break
    return result


async def update_routes(
    hub,
    ctx,
    route_table_id: str,
    old_routes: List = None,
    new_routes: List = None,
):
    """Update routes of a route table.

    This function compares the existing(old) routes of route table with new routes. Routes that are in the new route
    table but not in the old route table will be associated to route table. routes that are in the old route table  but
    not in the new route table will be disassociated from transit route table.

    Args:
        route_table_id(str):
            The AWS resource id of the existing route table

        old_routes(list):
            Routes of existing route table

        new_routes(list):
            New routes to update

    Returns:
        ``{"result": True|False, "comment": "A message", "ret": None}``

    """
    result = dict(comment=[], result=True, ret=None)

    # compare old_routes if routes are modified
    if new_routes is []:
        result["result"] = False
        result["comment"].append(
            "Route Table routes cannot be None. There should be at least one route in Route Table"
        )
        return result
    elif new_routes is not None:
        routes_to_modify = (
            hub.tool.aws.ec2.route_table.get_route_table_routes_modifications(
                old_routes, new_routes
            )
        )
        if not routes_to_modify["result"]:
            result["comment"] = routes_to_modify["comment"]
            result["result"] = False
            return result
        routes_to_add = routes_to_modify["routes_to_add"]
        routes_to_delete = routes_to_modify["routes_to_delete"]
        routes_to_replace = routes_to_modify["routes_to_replace"]

        if not ctx.get("test", False):
            if routes_to_delete:
                for route_to_delete in routes_to_delete:
                    ret = await hub.exec.boto3.client.ec2.delete_route(
                        ctx, RouteTableId=route_table_id, **route_to_delete
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + list(ret["comment"])
                        result["result"] = False
                        return result
                result["comment"].append(f"Deleted Routes: {routes_to_delete}")
            if routes_to_replace:
                for route_to_replace in routes_to_replace:
                    ret = await hub.exec.boto3.client.ec2.replace_route(
                        ctx, RouteTableId=route_table_id, **route_to_replace
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + list(ret["comment"])
                        result["result"] = False
                        return result
                result["comment"].append(f"Replace Routes: {routes_to_replace}")
            if routes_to_add:
                for route_to_add in routes_to_add:
                    ret = await hub.exec.boto3.client.ec2.create_route(
                        ctx, RouteTableId=route_table_id, **route_to_add
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + list(ret["comment"])
                        result["result"] = False
                        return result
                result["comment"].append(f"Added Routes: {routes_to_add}")
        if routes_to_add or routes_to_replace or routes_to_delete:
            result["comment"].append(f"Updated route table {route_table_id}")
    result["ret"] = {"routes": new_routes}
    return result
